'++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
'
'                    PC FAN Controler Soft for ATmega88
'
'                                  Rev.1.0
'
'                                                    2008.6 by T.Sugiyama
'
'++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

'$sim

$regfile = "m88def.dat"
$crystal = 16000000
$hwstack = 64
$swstack = 64
$framesize = 64

$baud = 9600

' AVR pin assign
'
'  1: PC6(RESET)   : ISP Reset
'  2: PD0(RxD)     : Serial UART
'  3: PD1(TxD)     : Serial UART
'  4: PD2(PCINT18) : LCD DB5
'  5: PD3(PCINT19) : FAN#4 Speed Sensor
'  6: PD4(PCINT20) : FAN#3 Speed Sensor
'  7: Vcc
'  8: GND
'  9: XTAL1
' 10: XTAL2
' 11: PD5(OC0B)    : FAN#3 PWM
' 12: PD6(OC0A)    : FAN#4 PWM
' 13: PD7(PCINT23) : FAN#2 Speed Sensor
' 14: PB0(PCINT0)  : FAN#1 Speed Sensor
' 15: PB1(OC1A)    : FAN#1 PWM
' 16: PB2(OC1B)    : FAN#2 PWM
' 17: PB3(MOSI)    : ISP MOSI
' 18: PB4(MISO)    : ISP MISO
' 19: PB5(SCK)     : ISP SCK
' 20: AVcc
' 21: Aref
' 22: AGND
' 23: PC0          : LCD DB6
' 24: PC1(PCINT9)  : LCD DB4
' 25: PC2(ADC2)    : Thermistor#1
' 26: PC3(ADC3)    : Thermistor#2
' 27: PC4(ADC4)    : Thermistor#3
' 28: PC5(ADC5)    : Thermistor#4

Config Pinb.1 = Output                                                          'FAN#1 PWM
Config Pinb.2 = Output                                                          'FAN#2 PWM
Config Pind.5 = Output                                                          'FAN#3 PWM
Config Pind.6 = Output                                                          'FAN#4 PWM

Config Pinb.0 = Input                                                           'FAN#1 Speed Sensor
Config Pind.7 = Input                                                           'FAN#2 Speed Sensor
Config Pind.4 = Input                                                           'FAN#3 Speed Sensor
Config Pind.3 = Input                                                           'FAN#4 Speed Sensor

PortB=&B00000001
PortD=&B10011000

Config Adc = Single , Prescaler = Auto

Config Timer1 = Pwm , Prescale = 1 , Clear Timer = 0 , Compare A Pwm = Clear Down , Compare B Pwm = Clear Down , Pwm = 8
Config Timer0 = Pwm , Prescale = 1 , Clear Timer = 0 , Compare A Pwm = Clear Down , Compare B Pwm = Clear Down

Config Timer2 = Timer , Prescale = 1024
On Timer2 Tmr2_int
On Pcint0 Pci_int
'On Pcint1 Pci_int
On pcint2 Pci_int

Config Serialin = Buffered , Size = 230

Dim Igr1 As Integer
Dim Igr2 As Integer
Dim Igr3 As Integer
Dim Igr4 As Integer

Dim Sgl1 As Single
Dim Sgl2 As Single
Dim Sgl3 As Single
Dim Sgl4 As Single

Dim Str1 As String * 10
Dim Str2 As String * 10
Dim Str3 As String * 10

Dim Message1 As String * 128
Dim StrArray(8) As string * 7

Dim Thm_chan As Byte
Dim Fan_chan As Byte
Dim Adc_chan As Byte

Dim Ratio As Word
Dim Chan  As Byte
Dim Tmpr  As Single

Dim Tmr_clock As Single             'CPU Clock (kHz)
Dim Tmr_cycle As Single             'Timer2 Cycle (ms)
Dim Tmr_over  As Single             'Timer2 Overflow time (ms)
Dim Tmr_over_cnt  As Byte           'Timer2 Overflow Count
Dim Tmr_over_cnt_max  As Byte       'Timer2 maximum Overflow Count (time out)
Dim Fan_cnt_time      As Single     'Fan Counted Time (ms)
Dim Fan_cnt_time2     As Single     'Fan Counted Time 2(ms)
Dim Fan_puls_min_wdth As Single     'Fan Pulse minimum width (ms)
Dim Fan_cnt       As Integer        'Fan Pulse Counter
Dim Fan_sns As bit                  'Fan sensor pulse value ( 0 or 1 )
Dim Intflag As Byte                 'Interrupt mode flag ( 0 or 1 )

Tmr_clock = 16000.0              'Crystal clock
Tmr_cycle = 1024.0/Tmr_clock     'Prescale = 1024
Tmr_over  = Tmr_cycle*256.0      'Timer2 is 8bit(256) Timer
Tmr_over_cnt_max = 20            'Timer2 maximum Overflow Count
Fan_puls_min_wdth = 2.0          'Fan Pulse minimum width (ms)

Dim Temper( 4 )     As Single
Dim Duty_rto( 6 )   As Single
Dim Tach( 6 )    As Integer

Dim Tmin( 4 )      As ERAM Integer  at 0
Dim Tmax( 4 )      As ERAM Integer  at 8
Dim Dmin( 4 )      As ERAM Integer  at 16
Dim Dmax( 4 )      As ERAM Integer  at 24
Dim FanThm( 4 )    As ERAM Integer  at 32
Dim Fanname( 6 )   As ERAM String*6 at 40
Dim Thmname( 4 )   As ERAM String*6 at 82

Declare Sub SndMsg( )
Declare Sub SndPrm( )
Declare Sub WrtPrm( )

Declare Function Temperature( Ratio As Word ) As Single
Declare Function Duty( Chan As Byte , Tmpr As Single ) As Integer
Declare Function Fanspeed( Chan AS Byte ) AS Integer
Declare Function Tmrpass( ) As Single

Dim Pmax( 4 )  As Integer
Pmax(1) = 255 : Pmax(2) = 255 : Pmax(3) = 255 : Pmax(4) = 255

'N@tXs[h
Pwm1a = Pmax(1)
Pwm1b = Pmax(2)
Pwm0b = Pmax(3)
Pwm0a = Pmax(4)
Waitms 3000

Enable Interrupts

'[v
Do


'  x
   For Thm_chan = 1 To 4

      Adc_chan = Thm_chan + 1
      Start Adc
      Ratio = Getadc( Adc_chan )
      Stop Adc
      Temper( Thm_chan ) = Temperature( Ratio )

      Sgl1 = Temper( Thm_chan )
      If Sgl1>=0.0 Then
         Str1 = Fusing( Sgl1, "#.#" )
      Else
         Str1 = "---"
      Endif
      Str2 = Thmname( Thm_chan )

      Print "Thm ," ; Thm_chan ; "," ; Str2  ; "," ; Str1

   Next Thm_chan

   Print

'  f[eBݒ
   For Fan_chan = 1 To 4

      Chan = FanThm( Fan_chan )
      Tmpr = Temper( Chan )
      Igr1 = Duty( Fan_chan , Tmpr )
      Igr2 = Igr1*100
      Duty_rto( Fan_chan ) = Igr2/Pmax( Fan_chan )


      Select Case Fan_chan
         Case 1 : Pwm1a = Igr1
         Case 2 : Pwm1b = Igr1
         Case 3 : Pwm0b = Igr1
         Case 4 : Pwm0a = Igr1
      End Select

   Next Fan_chan

'  ]
   For Fan_chan = 1 TO 4

      Igr1 = Fanspeed( Fan_chan )

'     ]O̎̏
      If Igr1 = 0 Then
         If Tach( Fan_chan ) > 0 Then
            Select Case Fan_chan
               Case 1 : Pwm1a = Pmax( 1 )
               Case 2 : Pwm1b = Pmax( 2 )
               Case 3 : Pwm0b = Pmax( 3 )
               Case 4 : Pwm0a = Pmax( 4 )
            End Select
            Waitms 1000
            Igr1 = Fanspeed( Fan_chan )
         Endif
      Endif

      Tach( Fan_chan ) = Igr1
      Str1 = Str( Igr1 )

      Sgl1 = Duty_rto( Fan_chan )
      Str2 = Fusing( Sgl1, "#.#" )

      Str3 = Fanname( Fan_chan )

      Print "Fan," ; Fan_chan ; "," ; Str3 ; "," ; Str1 ; "," ; Str2

   Next Fan_chan

   Print

   Call SndMsg()
   Print

   Call SndPrm()
   Print

Loop

End

$eeprom
$eepromhex

Tmindata:
   Data   25% ,  25% ,  25% ,  25%
Tmaxdata:
   Data   99% ,  99% ,  99% ,  99%
Dmindata:
   Data   20% ,  20% ,  20% ,  20%
Dmaxdata:
   Data  100% , 100% , 100% , 100%
FanThm:
   Data    1% ,   2% ,   3% ,   4%
Fanname:
   Data "Fan1  ","Fan2  ","Fan3  ","Fan4  ","Fan3+ ","Fan4+ "
Thmname:
   Data "Thm1  ","Thm2  ","Thm3  ","Thm4  "

$data


'***********************************************************
'T[~X^̓d牷xvZ֐

Function Temperature( Ratio As Word ) As Single
   'Temperature : xij
   'ratio       : T[~X^di0`1023j

   Local X As Single
   Local Y As Single

   X = Ratio
   Y = 1023.0 - X

   If Y <= 1.0 Then
      Temperature = -273.0
   Elseif Y >= 1022.0 Then
      Temperature = 999.0
   Else
      Y = X / Y
      X = Log(Y)
      Y = 0.29112 * X
      X = Y + 3.1359
      Y = 1000.0 / X
      Temperature = Y - 273.0
   End If

End Function

'***********************************************************
'T[~X^̉xFanDutyvZ֐

Function Duty( Fan_chan As Byte , Tmpr As Single ) As Integer
   'Ch      :  FAN Channel
   'Tmpr    :  xij

   Local D As Single

   Local W As Single
   Local X As Single
   Local Y As Single
   Local Z As Single

   Igr1 = Tmin( Fan_chan )
   Sgl1 = Igr1
   Igr2 = Tmax( Fan_chan )
   Sgl2 = Igr2
   Igr3 = Dmin( Fan_chan )
   Sgl3 = Igr3
   Igr4 = Dmax( Fan_chan )
   Sgl4 = Igr4

   If Tmpr <= Sgl1 Then
      D = Sgl3
   Elseif Tmpr => Sgl2 Then
      D = Sgl4
   Else
      W = Tmpr - Sgl1  'w=Tmpr-Tmin (deg)
      X = Sgl4 - Sgl3  'X=Dmax-Dmin (%)
      Y = Sgl2 - Sgl1  'Y=Tmax-Tmin (deg)
      Z = X / Y        'Z:Slope     (%/deg)
      Y = Z * W        'Y=slope*(Tmpr-Tmin)  (%)
      D = Y + Sgl3     'D=Y+Dmin    (%)
   End If

   D = D * Pmax( Fan_chan )
   D = D * 0.01
   Duty = Int( D )

End Function

'***********************************************************
'Fan̉]vvZ֐

Function Fanspeed( Fan_chan As Byte ) As Integer
   'Fan_chan   :  FAN Channel

   Local X As Single

   Fan_cnt = 0
   Fan_cnt_time = 0.0
   Tmr_over_cnt = 0
   Timer2 = 0

   If Fan_chan = 1 then

      PCmsk0 = &B00000001
      Enable PCINT0

   Else

      Select Case Fan_chan
         Case 2 :  PCmsk2 = &B10000000
         Case 3 :  PCmsk2 = &B00010000
         Case 4 :  PCmsk2 = &B00001000
      End Select
      Enable PCINT2

   Endif

   Enable Timer2
   Start Timer2

   Do
      If Tmr_over_cnt >= Tmr_over_cnt_max Then

         Stop Timer2
         Disable PCINT0
         Disable PCINT1
         Disable PCINT2
         Disable Timer2
         Goto Fan_speed_calc

      Endif
   Loop

Fan_speed_calc:

   Fan_cnt = Fan_cnt-1

   If Fan_cnt <= 0 Then

      Fanspeed = 0

   Else

      X = Fan_cnt  * 15000
      X = X / Fan_cnt_time
      Fanspeed = Int( X )

   Endif

End Function

'***********************************************************
'Timer2̌oߎԂvZ֐

Function Tmrpass( ) As Single
   'Tmr_over_cnt  :  Timer2 Overflow Count
   'Tmrpass       :  Timer2 passed Time (ms)

   local X As Single
   local Y As Single

   X = Tmr_over_cnt * Tmr_over
   Y = Timer2 * Tmr_cycle
   Tmrpass = X + Y

End Function

'***********************************************************

Sub SndMsg()

   Message1 = ""

   While Ischarwaiting() = 1                               'was there a char?

      Str1 = Inkey()
      If Str1 = Chr(&H0A) Then
'         Print "Msge,0," ;  Message1
         Call WrtPrm( )
         Message1 = ""
      Else
         Message1 = Message1 + Str1
      Endif

   Wend


End sub


'***********************************************************

Sub SndPrm()

      Print "Prm,Tmin" ;
      For Fan_chan = 1 To 4
         Igr1 = Tmin( Fan_chan )
         Str1 = Str( Igr1 )
         Print "," ; Str1 ;
      Next Fan_chan
      Print

      Print "Prm,Tmax" ;
      For Fan_chan = 1 To 4
         Igr1 = Tmax( Fan_chan )
         Str1 = Str( Igr1 )
         Print  "," ; Str1 ;
      Next Fan_chan
      Print

      Print "Prm,Dmin" ;
      For Fan_chan = 1 To 4
         Igr1 = Dmin( Fan_chan )
         Str1 = Str( Igr1 )
         Print "," ; Str1 ;
      Next Fan_chan
      Print

      Print "Prm,Dmax";
      For Fan_chan = 1 To 4
         Igr1 = Dmax( Fan_chan )
         Str1 = Str( Igr1 )
         Print "," ; Str1 ;
      Next Fan_chan
      Print

      Print "Prm,FanThm";
      For Fan_chan = 1 To 4
         Igr1 = FanThm( Fan_chan )
         Str1 = Str( Igr1 )
         Print "," ; Str1 ;
      Next Fan_chan
      Print

      Print "Prm,FanNam";
      For Fan_chan = 1 To 6
         Str1 = Fanname( Fan_chan )
         Print "," ; Str1 ;
      Next Fan_chan
      Print

      Print "Prm,ThmNam";
      For Thm_chan = 1 To 4
         Str1 = Thmname( Thm_chan )
         Print "," ; Str1 ;
      Next Thm_chan
      Print

End Sub

'***********************************************************

Sub WrtPrm()

      Igr1 = Split( Message1 , StrArray(1) , "," )
'      For Igr2 = 1 To Igr1
'         StrArray(Igr2) = Trim( StrArray(Igr2) )
'      Next Igr2

      If StrArray(1) = "SetPrm" Then

        Select Case StrArray(2)

         Case "Tmin"
            For Fan_chan = 1 To 4
               Igr2 = Fan_chan + 2
               Igr1 = Val( StrArray( Igr2 ))
               Tmin( Fan_chan ) = Igr1
            Next Fan_chan
         Case "Tmax"
            For Fan_chan = 1 To 4
               Igr2 = Fan_chan + 2
               Igr1 = Val( StrArray( Igr2 ))
               Tmax( Fan_chan ) = Igr1
            Next Fan_chan
         Case "Dmin"
            For Fan_chan = 1 To 4
               Igr2 = Fan_chan + 2
               Igr1 = Val( StrArray( Igr2 ))
               Dmin( Fan_chan ) = Igr1
            Next Fan_chan
         Case "Dmax"
            For Fan_chan = 1 To 4
               Igr2 = Fan_chan + 2
               Igr1 = Val( StrArray( Igr2 ))
               Dmax( Fan_chan ) = Igr1
            Next Fan_chan
         Case "FanThm"
            For Fan_chan = 1 To 4
               Igr2 = Fan_chan + 2
               Igr1 = Val( StrArray( Igr2 ))
               FanThm( Fan_chan ) = Igr1
            Next Fan_chan
         Case "FanNam"
            For Fan_chan = 1 To 6
               Igr2 = Fan_chan + 2
               Fanname( Fan_chan ) = StrArray( Igr2 )
            Next Fan_chan
         Case "ThmNam"
            For Thm_chan = 1 To 4
               Igr2 = Thm_chan + 2
               Thmname( Thm_chan ) = StrArray( Igr2 )
            Next Thm_chan

        End select

      Endif

End Sub

'***********************************************************
'PCINT0(Pcint0`7)荞݂̏
'PCINT2(Pcint16`23)荞݂̏
Pci_int:

   If Fan_cnt = 0 Then           'when measuring has not stated yet

         Timer2 = 0
         Tmr_over_cnt = 0
         Fan_cnt = 1
         Intflag = 1
         Fan_cnt_time = 0.0
         Fan_cnt_time2 = 0.0

   Elseif Fan_cnt >= 1 Then       'when measuring has already started

         sgl1 = Tmrpass( )
         sgl2 = sgl1-Fan_cnt_time

         If sgl2 > Fan_puls_min_wdth then

            Fan_cnt = Fan_cnt + 1
            Fan_cnt_time2 = Fan_cnt_time
            Fan_cnt_time = sgl1

         Else                     'soft lowpass filter

            Fan_cnt = Fan_cnt - 1
            Fan_cnt_time = Fan_cnt_time2 'rewind time

         Endif

   Endif

Return

'***********************************************************
'Timer2荞݂̏
Tmr2_int:

   Incr Tmr_over_cnt

Return